@@ -29,23 +29,6 @@
 
 static void ctdb_event_script_timeout(struct event_context *ev, struct timed_event *te, struct timeval t, void *p);
 
-/*
-  ctdbd sends us a SIGTERM when we should die.
- */
-static void sigterm(int sig)
-{
-	pid_t pid;
-
-	/* all the child processes will be running in the same process group */
-	pid = getpgrp();
-	if (pid == -1) {
-		kill(-getpid(), SIGKILL);
-	} else {
-		kill(-pid, SIGKILL);
-	}
-	_exit(1);
-}
-
 /* This is attached to the event script state. */
 struct event_script_callback {
 	struct event_script_callback *next, *prev;
@@ -62,7 +45,6 @@ struct ctdb_event_script_state {
 	struct event_script_callback *callback;
 	pid_t child;
 	int fd[2];
-	bool from_user;
 	enum ctdb_eventscript_call call;
 	const char *options;
 	struct timeval timeout;
@@ -118,10 +100,33 @@ int32_t ctdb_control_get_event_script_status(struct ctdb_context *ctdb,
 	return 0;
 }
 
-struct ctdb_script_tree_item {
-	const char *name;
-	int error;
-};
+/* To ignore directory entry return 0, else return non-zero */
+static int script_filter(const struct dirent *de)
+{
+	int namelen = strlen(de->d_name);
+
+	/* Ignore . and .. */
+	if (namelen < 3) {
+		return 0;
+	}
+
+	/* Skip temporary files left behind by emacs */
+	if (de->d_name[namelen-1] == '~') {
+		return 0;
+	}
+
+	/* Filename should start with [0-9][0-9]. */
+	if (!isdigit(de->d_name[0]) || !isdigit(de->d_name[1]) ||
+	    de->d_name[2] != '.') {
+		return 0;
+	}
+
+	if (namelen > MAX_SCRIPT_NAME) {
+		return 0;
+	}
+
+	return 1;
+}
 
 /* Return true if OK, otherwise set errno. */
 static bool check_executable(const char *dir, const char *name)
@@ -153,236 +158,129 @@ static bool check_executable(const char *dir, const char *name)
 
 static struct ctdb_scripts_wire *ctdb_get_script_list(struct ctdb_context *ctdb, TALLOC_CTX *mem_ctx)
 {
-	DIR *dir;
-	struct dirent *de;
-	struct stat st;
-	trbt_tree_t *tree;
+	struct dirent **namelist;
 	struct ctdb_scripts_wire *scripts;
-	TALLOC_CTX *tmp_ctx = talloc_new(ctdb);
-	struct ctdb_script_tree_item *tree_item;
 	int count;
 
-	/*
-	  the service specific event scripts 
-	*/
-	if (stat(ctdb->event_script_dir, &st) != 0 && 
-	    errno == ENOENT) {
-		DEBUG(DEBUG_CRIT,("No event script directory found at '%s'\n", ctdb->event_script_dir));
-		talloc_free(tmp_ctx);
-		return NULL;
-	}
-
-	/* create a tree to store all the script names in */
-	tree = trbt_create(tmp_ctx, 0);
-
 	/* scan all directory entries and insert all valid scripts into the 
 	   tree
 	*/
-	dir = opendir(ctdb->event_script_dir);
-	if (dir == NULL) {
-		DEBUG(DEBUG_CRIT,("Failed to open event script directory '%s'\n", ctdb->event_script_dir));
-		talloc_free(tmp_ctx);
+	count = scandir(ctdb->event_script_dir, &namelist, script_filter, alphasort);
+	if (count == -1) {
+		DEBUG(DEBUG_CRIT, ("Failed to read event script directory '%s' - %s\n",
+				   ctdb->event_script_dir, strerror(errno)));
 		return NULL;
 	}
 
-	count = 0;
-	while ((de=readdir(dir)) != NULL) {
-		int namlen;
-		unsigned num;
-
-		namlen = strlen(de->d_name);
-
-		if (namlen < 3) {
-			continue;
-		}
-
-		if (de->d_name[namlen-1] == '~') {
-			/* skip files emacs left behind */
-			continue;
-		}
-
-		if (de->d_name[2] != '.') {
-			continue;
-		}
-
-		if (sscanf(de->d_name, "%02u.", &num) != 1) {
-			continue;
-		}
-
-		if (strlen(de->d_name) > MAX_SCRIPT_NAME) {
-			DEBUG(DEBUG_ERR,("Script name %s too long! %u chars max",
-					 de->d_name, MAX_SCRIPT_NAME));
-			continue;
-		}
-
-		tree_item = talloc(tree, struct ctdb_script_tree_item);
-		if (tree_item == NULL) {
-			DEBUG(DEBUG_ERR, (__location__ " Failed to allocate new tree item\n"));
-			closedir(dir);
-			talloc_free(tmp_ctx);
-			return NULL;
-		}
-	
-		tree_item->error = 0;
-		if (!check_executable(ctdb->event_script_dir, de->d_name)) {
-			tree_item->error = errno;
-		}
-
-		tree_item->name = talloc_strdup(tree_item, de->d_name);
-		if (tree_item->name == NULL) {
-			DEBUG(DEBUG_ERR,(__location__ " Failed to allocate script name.\n"));
-			closedir(dir);
-			talloc_free(tmp_ctx);
-			return NULL;
-		}
-
-		/* store the event script in the tree */
-		trbt_insert32(tree, (num<<16)|count++, tree_item);
-	}
-	closedir(dir);
-
 	/* Overallocates by one, but that's OK */
-	scripts = talloc_zero_size(tmp_ctx,
+	scripts = talloc_zero_size(mem_ctx,
 				   sizeof(*scripts)
 				   + sizeof(scripts->scripts[0]) * count);
 	if (scripts == NULL) {
 		DEBUG(DEBUG_ERR, (__location__ " Failed to allocate scripts\n"));
-		talloc_free(tmp_ctx);
+		free(namelist);
 		return NULL;
 	}
 	scripts->num_scripts = count;
 
 	for (count = 0; count < scripts->num_scripts; count++) {
-		tree_item = trbt_findfirstarray32(tree, 1);
-
-		strcpy(scripts->scripts[count].name, tree_item->name);
-		scripts->scripts[count].status = -tree_item->error;
-
-		/* remove this script from the tree */
-		talloc_free(tree_item);
+		strcpy(scripts->scripts[count].name, namelist[count]->d_name);
+		scripts->scripts[count].status = 0;
+		if (!check_executable(ctdb->event_script_dir, namelist[count]->d_name)) {
+			scripts->scripts[count].status = -errno;
+		}
 	}
 
-	talloc_steal(mem_ctx, scripts);
-	talloc_free(tmp_ctx);
+	free(namelist);
 	return scripts;
 }
 
-static int child_setup(struct ctdb_context *ctdb)
+
+/* There cannot be more than 10 arguments to command helper. */
+#define MAX_HELPER_ARGS		(10)
+
+static bool child_helper_args(TALLOC_CTX *mem_ctx, struct ctdb_context *ctdb,
+			      enum ctdb_eventscript_call call,
+			      const char *options,
+			      struct ctdb_script_wire *current, int fd,
+			      int *argc, const char ***argv)
 {
-	if (setpgid(0,0) != 0) {
-		int ret = -errno;
-		DEBUG(DEBUG_ERR,("Failed to create process group for event scripts - %s\n",
-			 strerror(errno)));
-		return ret;
+	const char **tmp;
+	int n, i;
+	char *t, *saveptr, *opt;
+
+	tmp = talloc_array(mem_ctx, const char *, 10+1);
+	if (tmp == NULL)  goto failed;
+
+	tmp[0] = talloc_asprintf(tmp, "%d", fd);
+	tmp[1] = talloc_asprintf(tmp, "%s/%s", ctdb->event_script_dir, current->name);
+	tmp[2] = talloc_asprintf(tmp, "%s", ctdb_eventscript_call_names[call]);
+	n = 3;
+
+	/* Split options into individual arguments */
+	opt = talloc_strdup(mem_ctx, options);
+	if (opt == NULL) {
+		goto failed;
 	}
 
-	signal(SIGTERM, sigterm);
-	return 0;
-}
-
-static char *child_command_string(struct ctdb_context *ctdb,
-				       TALLOC_CTX *ctx,
-				       bool from_user,
-				       const char *scriptname,
-				       enum ctdb_eventscript_call call,
-				       const char *options)
-{
-	const char *str = from_user ? "CTDB_CALLED_BY_USER=1 " : "";
-
-	/* Allow a setting where we run the actual monitor event
-	   from an external source and replace it with
-	   a "status" event that just picks up the actual
-	   status of the event asynchronously.
-	*/
-	if ((ctdb->tunable.use_status_events_for_monitoring != 0)
-	    &&  (call == CTDB_EVENT_MONITOR)
-	    &&  !from_user) {
-		return talloc_asprintf(ctx, "%s%s/%s %s",
-				       str,
-				       ctdb->event_script_dir,
-				       scriptname, "status");
-	} else {
-		return talloc_asprintf(ctx, "%s%s/%s %s %s",
-				       str,
-				       ctdb->event_script_dir,
-				       scriptname,
-				       ctdb_eventscript_call_names[call],
-				       options);
-	}
-}
-
-static int child_run_one(struct ctdb_context *ctdb,
-			 const char *scriptname, const char *cmdstr)
-{
-	int ret;
-
-	ret = system(cmdstr);
-	/* if the system() call was successful, translate ret into the
-	   return code from the command
-	*/
-	if (ret != -1) {
-		ret = WEXITSTATUS(ret);
-	} else {
-		ret = -errno;
+	t = strtok_r(opt, " ", &saveptr);
+	while (t != NULL) {
+		tmp[n++] = talloc_strdup(tmp, t);
+		if (n > MAX_HELPER_ARGS) {
+			goto args_failed;
+		}
+		t = strtok_r(NULL, " ", &saveptr);
 	}
 
-	/* 127 could mean it does not exist, 126 non-executable. */
-	if (ret == 127 || ret == 126) {
-		/* Re-check it... */
-		if (!check_executable(ctdb->event_script_dir, scriptname)) {
-			DEBUG(DEBUG_ERR,("Script %s returned status %u. Someone just deleted it?\n",
-					 cmdstr, ret));
-			ret = -errno;
+	for (i=0; i<n; i++) {
+		if (tmp[i] == NULL) {
+			goto failed;
 		}
 	}
-	return ret;
-}
 
-/*
-  Actually run one event script
-  this function is called and run in the context of a forked child
-  which allows it to do blocking calls such as system()
- */
-static int child_run_script(struct ctdb_context *ctdb,
-			    bool from_user,
-			    enum ctdb_eventscript_call call,
-			    const char *options,
-			    struct ctdb_script_wire *current)
-{
-	char *cmdstr;
-	int ret;
-	TALLOC_CTX *tmp_ctx = talloc_new(ctdb);
+	/* Last argument should be NULL */
+	tmp[n++] = NULL;
 
-	ret = child_setup(ctdb);
-	if (ret != 0)
-		goto out;
+	*argc = n;
+	*argv = tmp;
+	return true;
 
-	cmdstr = child_command_string(ctdb, tmp_ctx, from_user,
-				      current->name, call, options);
-	CTDB_NO_MEMORY(ctdb, cmdstr);
 
-	DEBUG(DEBUG_DEBUG,("Executing event script %s\n",cmdstr));
+args_failed:
+	DEBUG(DEBUG_ERR, (__location__ " too many arguments '%s' to eventscript '%s'\n",
+			  options, ctdb_eventscript_call_names[call]));
 
-	if (current->status) {
-		ret = current->status;
-		goto out;
+failed:
+	if (tmp) {
+		talloc_free(tmp);
 	}
+	return false;
 
-	ret = child_run_one(ctdb, current->name, cmdstr);
-out:
-	talloc_free(tmp_ctx);
-	return ret;
 }
 
 static void ctdb_event_script_handler(struct event_context *ev, struct fd_event *fde,
 				      uint16_t flags, void *p);
 
+static const char *helper_prog = NULL;
+
 static int fork_child_for_script(struct ctdb_context *ctdb,
 				 struct ctdb_event_script_state *state)
 {
 	int r;
 	struct tevent_fd *fde;
 	struct ctdb_script_wire *current = get_current_script(state);
+	int argc;
+	const char **argv;
+	static const char *helper = BINDIR "/ctdb_event_helper";
+
+	if (helper_prog == NULL) {
+		const char *t = getenv("CTDB_EVENT_HELPER");
+		if (t != NULL) {
+			helper_prog = t;
+		} else {
+			helper_prog = helper;
+		}
+	}
 
 	current->start = timeval_current();
 
@@ -392,38 +290,32 @@ static int fork_child_for_script(struct ctdb_context *ctdb,
 		return -errno;
 	}
 
- 	if (!ctdb_fork_with_logging(state, ctdb, current->name, log_event_script_output,
-				    state, &state->child)) {
+	/* Arguments for helper */
+	if (!child_helper_args(state, ctdb, state->call, state->options, current,
+			       state->fd[1], &argc, &argv)) {
+		DEBUG(DEBUG_ERR, (__location__ " failed to create arguments for eventscript helper\n"));
+		r = -ENOMEM;
+		close(state->fd[0]);
+		close(state->fd[1]);
+		return r;
+	}
+
+	if (!ctdb_vfork_with_logging(state, ctdb, current->name,
+				     helper_prog, argc, argv,
+				     log_event_script_output,
+				     state, &state->child)) {
+		talloc_free(argv);
 		r = -errno;
 		close(state->fd[0]);
 		close(state->fd[1]);
 		return r;
 	}
 
-	/* If we are the child, do the work. */
-	if (state->child == 0) {
-		int rt;
-
-		debug_extra = talloc_asprintf(NULL, "eventscript-%s-%s:",
-					      current->name,
-					      ctdb_eventscript_call_names[state->call]);
-		close(state->fd[0]);
-		set_close_on_exec(state->fd[1]);
-		ctdb_set_process_name("ctdb_eventscript");
-
-		rt = child_run_script(ctdb, state->from_user, state->call, state->options, current);
-		/* We must be able to write PIPEBUF bytes at least; if this
-		   somehow fails, the read above will be short. */
-		write(state->fd[1], &rt, sizeof(rt));
-		close(state->fd[1]);
-		_exit(rt);
-	}
+	talloc_free(argv);
 
 	close(state->fd[1]);
 	set_close_on_exec(state->fd[0]);
 
-	DEBUG(DEBUG_DEBUG, (__location__ " Created PIPE FD:%d to child eventscript process\n", state->fd[0]));
-
 	/* Set ourselves up to be called when that's done. */
 	fde = event_add_fd(ctdb->ev, state, state->fd[0], EVENT_FD_READ,
 			   ctdb_event_script_handler, state);
@@ -552,44 +444,46 @@ static void ctdb_run_debug_hung_script(struct ctdb_context *ctdb, struct debug_h
 	int fd[2];
 	struct tevent_timer *ttimer;
 	struct tevent_fd *tfd;
+	const char **argv;
+	int i;
+
+	if (helper_prog == NULL) {
+		return;
+	}
 
 	if (pipe(fd) < 0) {
 		DEBUG(DEBUG_ERR,("Failed to create pipe fd for debug hung script\n"));
 		return;
 	}
 
-	if (!ctdb_fork_with_logging(ctdb, ctdb, "Hung script", NULL, NULL, &pid)) {
-		DEBUG(DEBUG_ERR,("Failed to fork a child process with logging to track hung event script\n"));
-		close(fd[0]);
-		close(fd[1]);
-		return;
-	}
-	if (pid == -1) {
-		DEBUG(DEBUG_ERR,("Fork for debug script failed : %s\n",
-				 strerror(errno)));
-		close(fd[0]);
-		close(fd[1]);
-		return;
-	}
-	if (pid == 0) {
-		char *buf;
+	argv = talloc_array(state, const char *, 5);
 
-		ctdb_set_process_name("ctdb_debug_hung_script");
-		if (getenv("CTDB_DEBUG_HUNG_SCRIPT") != NULL) {
-			debug_hung_script = getenv("CTDB_DEBUG_HUNG_SCRIPT");
+	argv[0] = talloc_asprintf(argv, "%d", fd[1]);
+	argv[1] = talloc_strdup(argv, debug_hung_script);
+	argv[2] = talloc_asprintf(argv, "%d", state->child);
+	argv[3] = talloc_strdup(argv, ctdb_eventscript_call_names[state->call]);
+	argv[4] = NULL;
+
+	for (i=0; i<4; i++) {
+		if (argv[i] == NULL) {
+			close(fd[0]);
+			close(fd[1]);
+			talloc_free(argv);
+			return;
 		}
-
-		close(fd[0]);
-
-		buf = talloc_asprintf(NULL, "%s %d %s",
-				      debug_hung_script, state->child,
-				      ctdb_eventscript_call_names[state->call]);
-		system(buf);
-		talloc_free(buf);
-
-		_exit(0);
 	}
 
+
+	if (!ctdb_vfork_with_logging(state, ctdb, "Hung-script",
+				     helper_prog, 5, argv, NULL, NULL, &pid)) {
+		DEBUG(DEBUG_ERR,("Failed to fork a child to track hung event script\n"));
+		talloc_free(argv);
+		close(fd[0]);
+		close(fd[1]);
+		return;
+	}
+
+	talloc_free(argv);
 	close(fd[1]);
 
 	ttimer = tevent_add_timer(ctdb->ev, state,
@@ -630,7 +524,6 @@ static void ctdb_event_script_timeout(struct event_context *ev, struct timed_eve
 	case CTDB_EVENT_RECOVERED:
 	case CTDB_EVENT_TAKE_IP:
 	case CTDB_EVENT_RELEASE_IP:
-	case CTDB_EVENT_STATUS:
 		state->scripts->scripts[state->current].status = 0;
 		DEBUG(DEBUG_ERR,("Ignoring hung script for %s call %d\n", state->options, state->call));
 		break;
@@ -698,6 +591,11 @@ static int event_script_destructor(struct ctdb_event_script_state *state)
 		status = 0;
 	}
 
+	state->ctdb->active_events--;
+	if (state->ctdb->active_events < 0) {
+		ctdb_fatal(state->ctdb, "Active events < 0");
+	}
+
 	/* This is allowed to free us; talloc will prevent double free anyway,
 	 * but beware if you call this outside the destructor!
 	 * the callback hangs off a different context so we walk the list
@@ -745,9 +643,7 @@ static bool check_options(enum ctdb_eventscript_call call, const char *options)
 	case CTDB_EVENT_START_RECOVERY:
 	case CTDB_EVENT_RECOVERED:
 	case CTDB_EVENT_MONITOR:
-	case CTDB_EVENT_STATUS:
 	case CTDB_EVENT_SHUTDOWN:
-	case CTDB_EVENT_RELOAD:
 	case CTDB_EVENT_IPREALLOCATED:
 		return count_words(options) == 0;
 
@@ -778,7 +674,6 @@ static int ctdb_event_script_callback_v(struct ctdb_context *ctdb,
 					const void *mem_ctx,
 					void (*callback)(struct ctdb_context *, int, void *),
 					void *private_data,
-					bool from_user,
 					enum ctdb_eventscript_call call,
 					const char *fmt, va_list ap)
 {
@@ -806,6 +701,14 @@ static int ctdb_event_script_callback_v(struct ctdb_context *ctdb,
 		}
 	}
 
+	/* Do not run new monitor events if some event is already running */
+	if (call == CTDB_EVENT_MONITOR && ctdb->active_events > 0) {
+		if (callback != NULL) {
+			callback(ctdb, -ECANCELED, private_data);
+		}
+		return 0;
+	}
+
 	/* Kill off any running monitor events to run this event. */
 	if (ctdb->current_monitor) {
 		struct ctdb_event_script_state *ms = talloc_get_type(ctdb->current_monitor, struct ctdb_event_script_state);
@@ -837,7 +740,6 @@ static int ctdb_event_script_callback_v(struct ctdb_context *ctdb,
 	state->callback->private_data = private_data;
 
 	state->ctdb = ctdb;
-	state->from_user = from_user;
 	state->call = call;
 	state->options = talloc_vasprintf(state, fmt, ap);
 	state->timeout = timeval_set(ctdb->tunable.script_timeout, 0);
@@ -867,12 +769,14 @@ static int ctdb_event_script_callback_v(struct ctdb_context *ctdb,
 	state->current = 0;
 	state->child = 0;
 
-	if (!from_user && (call == CTDB_EVENT_MONITOR || call == CTDB_EVENT_STATUS)) {
+	if (call == CTDB_EVENT_MONITOR) {
 		ctdb->current_monitor = state;
 	}
 
 	talloc_set_destructor(state, event_script_destructor);
 
+	ctdb->active_events++;
+
 	/* Nothing to do? */
 	if (state->scripts->num_scripts == 0) {
 		talloc_free(state);
@@ -906,7 +810,6 @@ int ctdb_event_script_callback(struct ctdb_context *ctdb,
 			       TALLOC_CTX *mem_ctx,
 			       void (*callback)(struct ctdb_context *, int, void *),
 			       void *private_data,
-			       bool from_user,
 			       enum ctdb_eventscript_call call,
 			       const char *fmt, ...)
 {
@@ -914,7 +817,7 @@ int ctdb_event_script_callback(struct ctdb_context *ctdb,
 	int ret;
 
 	va_start(ap, fmt);
-	ret = ctdb_event_script_callback_v(ctdb, mem_ctx, callback, private_data, from_user, call, fmt, ap);
+	ret = ctdb_event_script_callback_v(ctdb, mem_ctx, callback, private_data, call, fmt, ap);
 	va_end(ap);
 
 	return ret;
@@ -949,7 +852,7 @@ int ctdb_event_script_args(struct ctdb_context *ctdb, enum ctdb_eventscript_call
 
 	va_start(ap, fmt);
 	ret = ctdb_event_script_callback_v(ctdb, ctdb,
-			event_script_callback, &status, false, call, fmt, ap);
+			event_script_callback, &status, call, fmt, ap);
 	va_end(ap);
 	if (ret != 0) {
 		return ret;
@@ -1063,7 +966,7 @@ int32_t ctdb_run_eventscripts(struct ctdb_context *ctdb,
 
 	ret = ctdb_event_script_callback(ctdb, 
 			 state, run_eventscripts_callback, state,
-			 true, call, "%s", options);
+			 call, "%s", options);
 
 	if (ret != 0) {
 		ctdb_enable_monitoring(ctdb);
